(* Allows more print depth.                                         *)
Compiler.Control.Print.printDepth:= 100;

datatype arithExp = AST_ID of string | AST_NUM of int | 
    AST_NEG of arithExp | AST_PLUS of arithExp * arithExp | 
    AST_MINUS of arithExp * arithExp | AST_PRODUCT of arithExp * arithExp | 
    AST_QUOTIENT of arithExp * arithExp | AST_ERROR of string;
    
datatype token = ID of string | NUM of int | Plus | Minus | Mult | Div | Neg
  | LParen | RParen | EOF;

signature PCFLEXER =
sig
    val lex : string -> token list
    val lexstr : string -> token list
end


structure PCFlexer: PCFLEXER =
struct
    open TextIO;

    (* Return true if c is a letter or digit *)
    fun alphanum c = (Char.isAlpha c) orelse (Char.isDigit c)

    (* Extracts consecutive alphanumeric characters from the input to
       build up an identifier.  Returns a tuple containing the next
       identifier in the input and the input left over after removing
       the identifier.
       Precondition:  The initial character of the identifier has
         already been found and is passed in in the second parameter.
       Parameter 1:  Input to extract the identifier from
       Parameter 2:  The characters found so far in the identifier.
    *)
    fun getid nil id = (id, nil)
      | getid (s as c::rest) id = 
                if (alphanum c) then getid rest (id ^ (str(c)))
                else (id, s)

    (* Extracts consecutive digits from the input to
       build up an integer.  Returns a tuple containing the next
       integer in the input and the input left over after removing
       the integer.
       Precondition:  The initial digit of the integer has
         already been found and is passed in in the second parameter.
       Parameter 1:  Input to extract the integer from
       Parameter 2:  The digits found so far in the integer.
    *)
    fun getnum nil num = (num, nil)
      | getnum (s as c::rest) num =
                if (Char.isDigit c) then 
                    getnum rest (num*10 + ((ord c)-ord #"0"))
                else (num, s)
                    
    (* Return the list of tokens found in the input.
       Parameter:  A character list to tokenize
    *)
    fun gettokens nil = [EOF]
      | gettokens (#"+"::rest) = Plus::gettokens rest
      | gettokens (#"-"::rest) = Minus::gettokens rest
      | gettokens (#"*"::rest) = Mult::gettokens rest
      | gettokens (#"/"::rest) = Div::gettokens rest
      | gettokens (#"~"::rest) = Neg::gettokens rest
      | gettokens (#"("::rest) = LParen::gettokens rest
      | gettokens (#")"::rest) = RParen::gettokens rest
      | gettokens (c::rest) =
          if Char.isSpace c then
	    (* Recurse to skip white space *)
	    gettokens rest

	  else if Char.isAlpha c then
	    (* Return keyword or identifier *)
	    let val (id, remainder) = (getid rest (str c))
	    in
	      (ID id)::gettokens remainder
            end

          else if (Char.isDigit c) then
	    (* Return number *)
            let
	      val (num, remainder) = getnum rest ((ord c) - (ord #"0"))
	    in
	      (NUM (num))::gettokens remainder
	    end

          else
            (print ("Skipping illegal character "^(str c) ^"."); 
             gettokens rest)

    (* Returns the list of tokens found in a string. *)
    fun lexstr s = gettokens (explode s)
                
    (* Returns the list of tokens found in a file.
       Parameter:  filename *)
    fun lex file = 
        let
	    val strm = openIn file
            val filecontents = explode (input strm)
        in 
            (closeIn strm; gettokens filecontents)
        end
end;

(*  Now define the parsing part of the program, which takes a list of  *)
(*  tokens and returns an abstract syntax tree.                        *)

signature PCFPARSER =
sig
    val parse : token list -> arithExp
end

structure PCFparser : PCFPARSER =
struct
    (* Output an error message *)
    fun error (msg:string) = print msg

    (* Parses an expression.  If an expression is found, returns a
       tuple containing the ast for the expression and the input
       following the expression.  If an expression is not found, it
       returns an error and consumes all remaining input.
    *)
       (* Normal start to expression *)
    fun parseExpression ((ID v)::tl) = parseTermTail(parseTerm((ID v)::tl))
      | parseExpression ((NUM n)::tl) = parseTermTail(parseTerm((NUM n)::tl))
      | parseExpression (LParen::tl) = parseTermTail(parseTerm(LParen::tl))
      | parseExpression (Neg::tl) = parseTermTail(parseTerm(Neg::tl))
       (* error cases *)
      | parseExpression [EOF] = 
          (error "Unexpected end of input\n";
           (AST_ERROR "Unexpected end of input", nil))

      | parseExpression _ = 
          (error "Fatal error\n";
           (AST_ERROR "Fatal error", nil))

    (* Parses a term.  If a term is found, returns a
       tuple containing the ast for the term and the input
       following the factor.  If a term is not found, it
       returns an error and consumes all remaining input.
    *)
       (* Normal start to expression *)
and parseTerm ((ID v)::tl) = parseFactorTail(parseFactor((ID v)::tl))
      | parseTerm ((NUM n)::tl)  = parseFactorTail(parseFactor((NUM n)::tl))
      | parseTerm (LParen::tl)  = parseFactorTail(parseFactor(LParen::tl))
      | parseTerm (Neg::tl)  = parseFactorTail(parseFactor(Neg::tl))
       (* error cases *)
      | parseTerm [EOF] = 
          (error "Unexpected end of input\n";
           (AST_ERROR "Unexpected end of input", nil))

      | parseTerm _ = 
          (error "Fatal error\n";
           (AST_ERROR "Fatal error", nil))

    (* Parses a factor.  If a factor is found, returns a
       tuple containing the ast for the factor and the input
       following the factor.  If a factor is not found, it
       returns an error and consumes all remaining input.
    *)
       (* Normal start to factor *)
and parseFactor ((ID v)::rest) = (AST_ID v,rest)
      | parseFactor ((NUM n)::rest)  = (AST_NUM n,rest)
      | parseFactor (Neg::others)  = let
            val (fact,rest) = parseFactor others
           in (AST_NEG fact,rest)
           end
      | parseFactor (LParen::rest)  = 
            let val (eTree,rest2) = parseExpression rest
         in if hd rest2 = RParen then (eTree,tl rest2) 
		else (error "no closing parenthesis\n"; 
		(AST_ERROR "no closing parenthesis",nil))
	 end
       (* error cases *)
      | parseFactor [EOF] = 
          (error "Unexpected end of input\n";
           (AST_ERROR "Unexpected end of input", nil))

      | parseFactor _ = 
          (error "Fatal error\n";
           (AST_ERROR "Fatal error", nil))

    (* Parses the tail of a term.  If a term is found, returns a
       tuple containing the ast for the term and the input
       following the factor.  If a term is not found, it
       returns an error and consumes all remaining input.
    *)
       (* Normal start to expression *)
and parseTermTail (left, Plus::others) = let
               		val (right,rest) = parseTerm others
 		in parseTermTail(AST_PLUS (left,right), rest)
		end
      | parseTermTail (left, Minus::others) = let
               		val (right,rest) = parseTerm others
 		in parseTermTail(AST_MINUS (left,right), rest)
		end
      | parseTermTail (term, RParen::others) = (term, RParen::others)
      | parseTermTail (term, EOF::others) = (term, EOF::others)

       (* error cases *)
      | parseTermTail _ = 
          (error "Fatal error\n";
           (AST_ERROR "Fatal error", nil))

    (* Parses the tail of a factor.  If a factor is found, returns a
       tuple containing the ast for the factor and the input
       following the factor.  If a term is not found, it
       returns an error and consumes all remaining input.
    *)
       (* Normal start to expression *)
and parseFactorTail (left, Mult::others) = let
               		val (right,rest) = parseFactor others
 		in parseFactorTail(AST_PRODUCT(left,right), rest)
		end
      | parseFactorTail (left, Div::others) = let
               		val (right,rest) = parseFactor others
 		in parseFactorTail(AST_QUOTIENT (left,right), rest)
		end
       (* Legal tokens to follow TermTail *)
      | parseFactorTail (term, Plus::others) = (term, Plus::others)
      | parseFactorTail (term, Minus::others) = (term, Minus::others)
      | parseFactorTail (term, RParen::others) = (term, RParen::others)
      | parseFactorTail (term, EOF::others) = (term, EOF::others)

       (* error cases *)
      | parseFactorTail _ = 
          (error "Fatal error\n";
           (AST_ERROR "Fatal error", nil))



    (* Return an AST for the list of tokens passed in. *)
  fun parse tokens =
      let 
        val (ast1, rest) = parseExpression tokens
      in
        if rest = [EOF] orelse rest = nil then
          ast1
        else
          (error "Fatal error -- more input than expected.\n";
           AST_ERROR "More input than expected.")
      end

end;  (* of structure PCFParser *)


(*  #CheckMatch Debug := true; -- put back in if run on Mac *)

(* The final definition of the function puts the pieces together.   *)

fun parsefile str = PCFparser.parse (PCFlexer.lex str);

fun parsestr str = PCFparser.parse (PCFlexer.lexstr str)
